from typing import *
from dataclasses import dataclass, field
from functools import cached_property
from typing import Any

from wcd_envoyer.utils import importable_prop, KwargsInjector

if TYPE_CHECKING:
    from wcd_envoyer.models import Message, Template, ChannelConfig


__all__ = (
    'Recipient', 'Recipients',  'TemplatedMessage', 'TemplatedMessages',
    'FailedMessages', 'SucceededFailedMessages',
    'MessageData',
    'BaseMessagesMaker', 'MessagesMaker', 'SplitRecipientsMessagesMaker',
    'MultiRecipientsMessagesMakerMixin', 'SplitRecipientsMessagesMakerMixin',
    'BaseMessagingBackend',
)

Recipient = Dict[str, Any]
Recipients = Sequence[Recipient]
TemplatedMessage = Tuple['MessageData', 'Template']
TemplatedMessages = List[TemplatedMessage]
FailedMessages = List[Tuple['Message', Any]]
SucceededFailedMessages = Tuple[List['Message'], FailedMessages]


@dataclass
class MessageData:
    channel: str
    event: str
    recipients: Recipients
    context: dict = field(default_factory=dict)


class BaseMessagesMaker:
    def __init__(self, backend, recipients_resolver, template_renderer):
        self.recipients_resolver = recipients_resolver
        self.template_renderer = template_renderer

    def normalize_message_data(
        self,
        config: 'ChannelConfig',
        message_data: MessageData,
        context: dict = {},
    ):
        return message_data
        for recipient in self.recipients_resolver(message_data.recipients):
            yield MessageData(
                message_data.channel,
                message_data.event,
                recipients=[recipient],
                context=message_data.context,
            )

    def make_message(
        self,
        message_data: MessageData,
        template: 'Template',
        context: dict = {}
    ) -> 'Message':
        # FIXME: Circular import...

        from wcd_envoyer.models import Message
        data = {**context, **message_data.context}
        rendered = self.template_renderer(template, context=data)

        return Message(
            channel=message_data.channel,
            event=message_data.event,
            recipients=message_data.recipients,
            data={**data, **rendered}
        )

    def __call__(
        self,
        config: 'ChannelConfig',
        templated_messages: TemplatedMessages,
        context: dict = {},
    ) -> List['Message']:
        messages = []

        for message_data, template in templated_messages:
            for data in self.normalize_message_data(
                config, message_data, context=context,
            ):
                messages.append(self.make_message(data, template, context))

        return messages


class MultiRecipientsMessagesMakerMixin:
    def normalize_message_data(
        self,
        config: 'ChannelConfig',
        message_data: MessageData,
        context: dict = {},
    ):
        recipients = self.recipients_resolver(message_data.recipients)

        if len(recipients) == 0:
            return

        yield MessageData(
            message_data.channel,
            message_data.event,
            recipients=recipients,
            context=message_data.context,
        )


class SplitRecipientsMessagesMakerMixin:
    def normalize_message_data(
        self,
        config: 'ChannelConfig',
        message_data: MessageData,
        context: dict = {},
    ):
        for recipient in self.recipients_resolver(message_data.recipients):
            yield MessageData(
                message_data.channel,
                message_data.event,
                recipients=[recipient],
                context=message_data.context,
            )


class MessagesMaker(MultiRecipientsMessagesMakerMixin, BaseMessagesMaker):
    pass


class SplitRecipientsMessagesMaker(
    SplitRecipientsMessagesMakerMixin, BaseMessagesMaker,
):
    pass


class BaseMessagingBackend(KwargsInjector):
    # FIXME: A bit shitty API here:
    recipient_resolver = importable_prop(lambda x: x)
    config_form_class = importable_prop(
        'wcd_envoyer.channels.forms.BaseConfigForm',
    )
    template_form_class = importable_prop(
        'wcd_envoyer.channels.forms.BaseTemplateForm',
    )
    template_renderer = importable_prop(
        'wcd_envoyer.channels.renderers.django_template_renderer',
    )
    messages_maker_class: Type[BaseMessagesMaker] = importable_prop(
        'wcd_envoyer.channels.backend.MessagesMaker',
    )
    messages_maker: BaseMessagesMaker = cached_property(
        lambda self: self.messages_maker_class(
            self,
            recipients_resolver=self.resolve_recipients,
            template_renderer=self.render_template,
        )
    )

    def resolve_recipients(self, recipients: Recipients):
        resolver = self.recipient_resolver
        resolved = (resolver(recipient) for recipient in recipients)

        return [x for x in resolved if x is not None]

    # TODO: Move template related methods into different class.
    def render_template(self, template: 'Template', context: dict = {}):
        # TODO: Change the way it handles translation
        initial_data = template.data
        template.data = template.translated.data
        renderer = self.template_renderer
        form = self.template_form_class(instance=template)
        data = {**form.initial}

        for field in form.get_renderable_fields():
            data[field] = renderer(data.get(field, ''), {**context, **data})

        template.data = initial_data

        return data

    def make_messages(
        self,
        config: 'ChannelConfig',
        templated_messages: TemplatedMessages,
        context: dict = {},
    ) -> List['Message']:
        return self.messages_maker(config, templated_messages, context=context)

    def send(
        self,
        config: 'ChannelConfig',
        messages: Sequence['Message'],
        context: dict = {},
    ) -> SucceededFailedMessages:
        raise NotImplementedError()



from dataclasses import dataclass, field
from typing import *

from px_settings.contrib.django import settings as setting_wrap

from .channels import ChannelInfoType
from .events import EventInfoType
from .const import SETTINGS_PREFIX


__all__ = 'Settings', 'settings',


@setting_wrap(SETTINGS_PREFIX)
@dataclass
class Settings:
    CHANNELS: Sequence[ChannelInfoType] = field(default_factory=list)
    EVENTS: Sequence[EventInfoType] = field(default_factory=list)
    JSON_ENCODER: str = 'django.core.serializers.json.DjangoJSONEncoder'


settings = Settings()


# class KLD_LETTRE_CHANNELS(TextChoices):
#     email = 'email', pgettext_lazy('kld_lettre', 'Email')
#     some_channel = 'sendpulse', pgettext_lazy('kld_lettre', 'Some channel')


# class KLD_LETTRE_EVENTS(TextChoices):
#     event1 = 'event1', pgettext_lazy('kld_lettre', 'Event 1')
#     event2 = 'event2', pgettext_lazy('kld_lettre', 'Event 2')


# KLD_LETTRE_EVENTS_CONTEXT = {
#     KLD_LETTRE_EVENTS.event1: {
#         'var1': pgettext_lazy('kld_lettre', 'Var 1'),
#         'var2': pgettext_lazy('kld_lettre', 'Var 2'),
#     },
#     KLD_LETTRE_EVENTS.event2: {
#         'var2': pgettext_lazy('kld_lettre', 'Var 2'),
#         'var3': pgettext_lazy('kld_lettre', 'Var 3'),
#     },
# }

# KLD_LETTRE_CHANNEL_EVENTS = {
#     KLD_LETTRE_CHANNELS.email: [
#         KLD_LETTRE_EVENTS.event1,
#         KLD_LETTRE_EVENTS.event2,
#     ],
#     KLD_LETTRE_CHANNELS.some_channel: [
#         KLD_LETTRE_EVENTS.event1,
#     ]
# }

# KLD_LETTRE_GENERATOR_RENDERERS = {
#     'kld_lettre.renderers.Jinja2Renderer': (
#         'kld_lettre.generators.field_plug.Jinja2PlugGenerator'
#     ),
# }

# KLD_LETTRE_BACKENDS = {
#     KLD_LETTRE_CHANNELS.email: (
#         'kld_lettre.backends.email.EmailMessageBackend'
#     ),
#     KLD_LETTRE_CHANNELS.some_channel: (
#         'kld_lettre.backends.base.BaseMessageBackend',
#         {
#             'config_form_path': 'kld_lettre.forms.BaseConfigForm',  # default value
#             'recipients_resolver_path': (
#                 'kld_lettre.recipients.BaseRecipientsResolver'  # default value
#             ),
#             'template_form_path': 'kld_lettre.forms.BaseTemplateForm',  # default value
#             'template_renderer_path': 'kld_lettre.renderers.Jinja2Renderer',  # default value
#             'template_data_generator_path': (
#                 'kld_lettre.generators.template_data.BaseTemplateDataGenerator'  # default value
#             ),
#         }
#     ),
# }



from django.contrib import admin
from django.utils.html import mark_safe
from django.utils.translation import pgettext_lazy, pgettext

from pxd_lingua.admin import TranslationsInlineAdmin

from wcd_envoyer import events
from wcd_envoyer.channels.forms import BaseTemplateForm
from wcd_envoyer.models import Template, TemplateTranslation
from wcd_envoyer.shortcuts import get_backend

from .forms import AddTemplateForm
from .utils import (
    EntangledFormSet, FormAdminMixin, JsonDataAdminMixin, render_html_table
)


class TemplateTranslationsInlineAdmin(TranslationsInlineAdmin):
    formset = EntangledFormSet
    model = TemplateTranslation

    def get_ent_form(self, parent_obj):
        backend = get_backend(parent_obj.channel)

        BaseForm = (
            BaseTemplateForm
            if backend is None else
            backend.template_form_class
        )

        trans_untangled_fields = [
            field for field in getattr(BaseForm.Meta, 'untangled_fields', [])
            if field not in {'channel', 'event', 'is_active'}
        ] + ['language']

        class TransForm(BaseForm, self.form):
            class Meta(BaseForm.Meta):
                fields = 'language',
                model = self.model
                untangled_fields = trans_untangled_fields

        return TransForm

    def get_formset(self, request, obj=None, **kwargs):
        self.form = self.__class__.form

        if obj:
            if not hasattr(self, '_ent_form'):
                self._ent_form = self.get_ent_form(obj)
            self.form = self._ent_form or self.form
        return super().get_formset(request, obj=obj, **kwargs)


@admin.register(Template)
class TemplateAdmin(JsonDataAdminMixin, FormAdminMixin, admin.ModelAdmin):
    add_form = AddTemplateForm
    add_fieldsets = (None, {'fields': ['channel', 'event']}),
    add_inlines = ()
    inlines = TemplateTranslationsInlineAdmin,
    list_display = 'channel', 'event', 'is_active', 'created_at', 'updated_at',
    list_filter = 'channel', 'event', 'is_active',
    date_hierarchy = 'created_at'
    readonly_fields = 'legend', 'json_data',
    search_fields = 'channel', 'event', 'data', 'status',

    def get_backend_form_class(self, request, obj, backend):
        return backend.template_form_class

    def legend(self, obj):
        if not getattr(obj, 'pk', None):
            return '-'

        event = events.registry.get(obj.event)

        if event is None:
            return '-'

        data = [(
            pgettext('wcd_envoyer', 'Key'),
            pgettext('wcd_envoyer', 'Title'),
            pgettext('wcd_envoyer', 'Description'),
        )] + [
            (key, str(title), str(description))
            for key, title, description in event.context
        ]

        return mark_safe(render_html_table(data))
    legend.short_description = pgettext_lazy('wcd_envoyer', 'Legend')



from typing import *
from functools import lru_cache, cached_property
from itertools import groupby

from django.utils.module_loading import import_string


__all__ = (
    'Numeric',
    'cached_import_string',
    'sort_and_group',
    'importable_prop',
    'KwargsInjector',
)

T = TypeVar('T')
Numeric = Union[SupportsInt, SupportsFloat]


@lru_cache
def cached_import_string(path: Optional[str] = None):
    if path is None:
        return None

    return import_string(path)


def sort_and_group(
    items: Iterable[T],
    key: Optional[Callable[[T], Any]] = None,
):
    return groupby(sorted(items, key=key), key=key)


def autoimport(value):
    return cached_import_string(value) if isinstance(value, str) else value


class importable_prop(cached_property):
    def __init__(self, default: T = None):
        super().__init__(lambda _: autoimport(default))

    def __set__(self, instance, value: Any) -> None:
        cache = instance.__dict__
        cache[self.attrname] = autoimport(value)


class KwargsInjector:
    def __init__(self, **kwargs):
        cls = self.__class__

        for key, value in kwargs.items():
            if not hasattr(cls, key):
                raise TypeError(
                    "%s() received an invalid keyword %r."
                    "only accepts arguments that are already "
                    "attributes of the class." % (cls.__name__, key)
                )

            if value is None:
                continue

            setattr(self, key, value)



from django.utils.translation import pgettext_lazy
from wcd_notifications.compat import TextChoices


__all__ = 'SETTINGS_PREFIX', 'LETTER_STATUSES',

SETTINGS_PREFIX = 'WCD_ENVOYER'


class MessageStatus(TextChoices):
    NEW = '010-new', pgettext_lazy('wcd_envoyer', 'New')
    PENDING = '030-pending', pgettext_lazy('wcd_envoyer', 'Pending')
    SENT = '060-sent', pgettext_lazy('wcd_envoyer', 'Sent')
    FAILED = '080-failed', pgettext_lazy('wcd_envoyer', 'Failed')



from django.db import models
from django.utils.translation import pgettext_lazy

from wcd_notifications.compat import JSONField
from pxd_lingua import create_translated_model

from .. import channels, events
from ..utils import DynamicChoicesField, get_json_encoder
from ..const import MessageStatus
from .utils import DateTimeModel


__all__ = 'Message', 'Template', 'TemplateTranslation',


class Template(DateTimeModel):
    class Meta:
        unique_together = ('channel', 'event'),
        verbose_name = pgettext_lazy('wcd_envoyer', 'Template')
        verbose_name_plural = pgettext_lazy('wcd_envoyer', 'Templates')

    channel = DynamicChoicesField(
        pgettext_lazy('wcd_envoyer', 'Channel'),
        max_length=512, choices=channels.registry.get_choices,
    )
    event = DynamicChoicesField(
        pgettext_lazy('wcd_envoyer', 'Event'),
        max_length=512, choices=events.registry.get_choices,
    )
    data = JSONField(
        pgettext_lazy('wcd_envoyer', 'Data'), default=dict,
        blank=True, null=False, encoder=get_json_encoder(),
    )

    is_active = models.BooleanField(
        pgettext_lazy('wcd_envoyer', 'Is active'), default=False,
    )

    def __str__(self) -> str:
        return pgettext_lazy('wcd_envoyer', '{channel}: {event}').format(
            channel=self.get_channel_display(),
            event=self.get_event_display(),
        )


class Message(DateTimeModel):
    Status = MessageStatus

    class Meta:
        verbose_name = pgettext_lazy('wcd_envoyer', 'Message')
        verbose_name_plural = pgettext_lazy('wcd_envoyer', 'Messages')
        ordering = '-created_at',

    status = models.CharField(
        pgettext_lazy('wcd_envoyer', 'Status'),
        max_length=512, default=MessageStatus.NEW,
        choices=MessageStatus.choices,
    )
    channel = DynamicChoicesField(
        pgettext_lazy('wcd_envoyer', 'Channel'),
        max_length=512, choices=channels.registry.get_choices,
    )
    event = DynamicChoicesField(
        pgettext_lazy('wcd_envoyer', 'Event'),
        max_length=512, choices=events.registry.get_choices,
    )
    recipients = JSONField(
        pgettext_lazy('wcd_envoyer', 'Recipients'),
        default=list, blank=False, null=False,
    )
    data = JSONField(
        pgettext_lazy('wcd_envoyer', 'Data'), default=dict,
        blank=True, null=False, encoder=get_json_encoder(),
    )

    def __str__(self) -> str:
        return pgettext_lazy('wcd_envoyer', 'Message #{id}: {status}').format(
            id=self.pk, status=self.get_status_display(),
        )


TemplateTranslation = create_translated_model(
    Template, fields=('data',)
)



from django.contrib import admin
from django.contrib.messages import INFO

from wcd_envoyer.models import Message
from wcd_envoyer.shortcuts import default_sender
from django.utils.translation import pgettext, pgettext_lazy

from .utils import JsonDataAdminMixin


@admin.register(Message)
class MessageAdmin(JsonDataAdminMixin, admin.ModelAdmin):
    messages_sender = default_sender
    actions = 'resend_action',
    list_display = 'channel', 'event', 'status', 'created_at', 'updated_at',
    list_filter = 'channel', 'event', 'status',
    readonly_fields = 'json_data', 'created_at', 'updated_at',
    date_hierarchy = 'created_at'
    search_fields = 'channel', 'event', 'recipients', 'data', 'status',

    def resend_action(self, request, qs):
        messages = list(qs)
        self.messages_sender.resend(messages)

        self.message_user(
            request=request,
            message=(
                pgettext('wcd_envoyer', 'Resent {} letters.')
                .format(len(messages))
            ),
            level=INFO,
        )
    resend_action.short_description = pgettext_lazy(
        'wcd_envoyer', 'Resend letters',
    )



from typing import *
import json

from django.forms import BaseInlineFormSet
from django.forms.formsets import DELETION_FIELD_NAME
from django.utils.html import mark_safe
from django.utils.translation import pgettext_lazy

from wcd_envoyer.shortcuts import get_backend


__all__ = (
    'EntangledFormSet',
    'FormAdminMixin', 'JsonDataAdminMixin',
    'prettify_json', 'render_html_table',
)


class EntangledFormSet(BaseInlineFormSet):
    def add_fields(self, form, index):
        super().add_fields(form, index)

        if self.can_delete:
            form._meta.untangled_fields.append(DELETION_FIELD_NAME)


class FormAdminMixin:
    add_form = None
    add_fieldsets = None
    add_inlines = None

    def get_backend_form_class(self, request, obj, backend):
        raise NotImplementedError()

    def get_form(self, request, obj=None, **kwargs):
        form = None

        if obj is None and self.add_form is not None:
            form = self.add_form

        if obj is not None:
            backend = get_backend(obj.channel)

            if backend is not None:
                form = self.get_backend_form_class(request, obj, backend)

        if form is not None:
            kwargs['form'] = form

        return super().get_form(request, obj, **kwargs)

    def get_fieldsets(self, request, obj=None):
        self.inlines = self.__class__.inlines

        if obj is None and self.add_inlines is not None:
            self.inlines = self.add_inlines

        if obj is None and self.add_fieldsets is not None:
            return self.add_fieldsets

        return super().get_fieldsets(request, obj)


class JsonDataAdminMixin:
    def json_data(self, obj):
        return mark_safe(prettify_json(getattr(obj, 'data', {})))
    json_data.short_description = pgettext_lazy('wcd_envoyer', 'Json data')


def render_html_table(data: List[Tuple]) -> str:
    html = '<table style="border: 1px solid #ecf2f6;"><tbody>'
    for line in data:
        html += '<tr><td>'
        html += '</td><td>'.join(line)
        html += '</td></tr>'
    html += '</tbody></table>'
    return html


def prettify_json(data: Dict) -> str:
    return ''.join([
        '<span style="white-space: pre-wrap; font-family: monospace">',
            json.dumps(data, ensure_ascii=False, indent=4),
        '</span>',
    ])



from collections import OrderedDict


__all__ = 'Registry', 'EntryRegistry',


class Registry(OrderedDict):
    def add(self, key, value):
        assert key not in self, f'{key} already registered.'

        self[key] = value

    register = add


class EntryRegistry(Registry):
    def transform_input(self, entry):
        raise NotImplementedError()

    def add(self, entry):
        key, value = self.transform_input(entry)
        return super().add(key, value)

    register = add



from typing import *

from wcd_envoyer.models import Message
from wcd_envoyer.channels import MessageData
from wcd_envoyer.services.sender import Sender


class CelerySender(Sender):
    celery_task_caller: Callable = None

    def send_messages_through_celery(self, message_ids, context):
        Message.objects.filter(pk__in=message_ids).update(
            status=Message.Status.PENDING
        )

        return self.celery_task_caller(
            message_ids=message_ids, context=context, only_pending=True,
        )

    def send(
        self,
        messages_data: Sequence[MessageData],
        context: dict = {},
        only_channels: Optional[Sequence[str]] = None,
    ):
        messages_groups = self.prepare_messages(
            messages_data, context=context, only_channels=only_channels,
        )

        if len(messages_groups) != 0:
            self.send_messages_through_celery(
                [y.pk for _, _, x in messages_groups for y in x], context,
            )

        return [], []

    def resend(self, messages: Iterable[Message], context: dict = {}):
        self.send_messages_through_celery([x.pk for x in messages], context)
        return [], []



from typing import *
from django import forms

from django.contrib.postgres.forms import SimpleArrayField

from django.conf import settings
from django.core.mail import EmailMultiAlternatives, get_connection
from django.utils.translation import pgettext_lazy

from wcd_envoyer.channels.backend import (
    BaseMessagingBackend, BaseMessagesMaker, SplitRecipientsMessagesMaker,
)
from wcd_envoyer.channels.forms import BaseTemplateForm, BaseConfigForm
from wcd_envoyer.utils import importable_prop
from wcd_envoyer.models import Message, ChannelConfig


__all__ = 'SendMailTemplateForm', 'SendMailBackend',


class SendMailConfigForm(BaseConfigForm):
    class Meta(BaseConfigForm.Meta):
        entangled_fields = {'data': ['from_email', 'reply_to']}

    from_email = forms.CharField(
        label=pgettext_lazy('wcd_envoyer', 'From email'),
        initial='', required=False,
    )
    reply_to = SimpleArrayField(
        base_field=forms.CharField(), delimiter=',',
        label=pgettext_lazy('wcd_envoyer', 'Reply to'),
        initial=[], required=False,
    )


class SendMailTemplateForm(BaseTemplateForm):
    renderable_fields = ['subject', 'plain_text', 'content']

    class Meta(BaseTemplateForm.Meta):
        entangled_fields = {'data': ['subject', 'plain_text', 'content']}

    subject = forms.CharField(
        label=pgettext_lazy('wcd_envoyer', 'Subject'),
    )
    content = forms.CharField(
        label=pgettext_lazy('wcd_envoyer', 'Content'),
        widget=forms.Textarea(),
    )
    plain_text = forms.CharField(
        label=pgettext_lazy('wcd_envoyer', 'Plain text'),
        widget=forms.Textarea(),
        initial='', required=False,
    )


class SendMailBackend(BaseMessagingBackend):
    content_type = 'text/html'
    recipient_resolver = importable_prop(lambda x: x.get('email'))
    config_form_class = importable_prop(SendMailConfigForm)
    template_form_class = importable_prop(SendMailTemplateForm)
    messages_maker_class: Type[BaseMessagesMaker] = importable_prop(
        SplitRecipientsMessagesMaker
    )

    def get_default_from_email(self):
        if 'des' in settings.INSTALLED_APPS:
            from des.models import DynamicEmailConfiguration
            return DynamicEmailConfiguration.get_solo().from_email

        return settings.DEFAULT_FROM_EMAIL

    def create_email(
        self,
        config: ChannelConfig,
        message: Message,
        context: dict = {}
    ) -> EmailMultiAlternatives:
        data = message.data
        config_data = config.data
        content = data.get('content')
        alternatives = []

        if content:
            alternatives.append((content, self.content_type))

        return EmailMultiAlternatives(
            subject=data.get('subject', ''),
            body=data.get('plain_text', ''),
            from_email=(
                config_data.get('from_email', None)
                or
                self.get_default_from_email()
                or
                None
            ),
            to=message.recipients,
            alternatives=alternatives,
            reply_to=config_data.get('reply_to', None) or None,
        )

    def create_emails(
        self,
        config: ChannelConfig,
        messages: Sequence[Message],
        context: dict = {}
    ) -> List[Tuple[Message, EmailMultiAlternatives]]:
        return [
            (message, self.create_email(config, message, context))
            for message in messages
        ]

    def send(
        self,
        config: ChannelConfig,
        messages: Sequence[Message],
        context: dict = {},
    ):
        emails = self.create_emails(config, messages, context=context)
        sent, failed = [], []

        try:
            with get_connection(fail_silently=False) as connection:
                for message, email in emails:
                    try:
                        count = connection.send_messages([email])

                        if count == 1:
                            sent.append(message)
                        else:
                            failed.append((message, None))
                    except Exception as e:
                        failed.append((message, e))

        except Exception as e:
            return [], [(x, e) for x in messages]

        return sent, failed
